import argparse
import json
import logging
import os
import re
import shutil
import sys
from collections import defaultdict
from enum import Enum
from os.path import expanduser

import paramiko
import yaml

import utils
from visualize_task_json import draw_tasks_dag

HOME_DIR = expanduser("~")
FLASHPROF_DIR = os.path.join(os.path.realpath('.'), 'flashprof')
FLASHPROF_CLUSTER_DIR = os.path.join(FLASHPROF_DIR, 'cluster')


class RENDER_TYPE(Enum):
    TASK_DAG = 'task_dag'

    def __str__(self):
        return self.value


def _get_tiup_config(cluster_name):
    cluster_dir = '{}/.tiup/storage/cluster/clusters/{}/'.format(HOME_DIR, cluster_name)
    ssh_key_file = '{}/ssh/id_rsa'.format(cluster_dir)
    logging.debug('cluster_dir = {}'.format(cluster_dir))
    meta_filename = '{}/meta.yaml'.format(cluster_dir)
    fd = open(meta_filename, 'r')
    meta = yaml.safe_load(fd.read())
    fd.close()
    username = meta['user']
    tiflash_servers = meta['topology']['tiflash_servers']
    return username, ssh_key_file, tiflash_servers


def _copy_log_file(host, port, username, ssh_key_file, remote_log_dir, cluster_name):
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(host, port, username, key_filename=ssh_key_file)
    # sftp = ssh.open_sftp()
    remote_log_filename = os.path.join(remote_log_dir, 'tiflash.log')
    local_log_filename = os.path.join(FLASHPROF_CLUSTER_DIR, cluster_name, 'log', '{}.tiflash.log'.format(host))
    # remote_log_filename = '/mnt/pingcap/tiflash_mpp_profiler/log/172.16.5.59.tiflash.log'
    command = r'grep -P "\[\"MPPTask:<query:\d+,task:\d+> mpp_task_tracing" {}'.format(remote_log_filename)
    logging.debug('executing ssh command: {}'.format(command))
    stdin, stdout, stderr = ssh.exec_command(command)
    logging.info('grep & scp {}@{}:{}/{} {}'.format(username, host, port, remote_log_filename, local_log_filename))
    # sftp.get(remote_log_filename, local_log_filename)
    with open(local_log_filename, 'wb') as fd:
        fd.write(stdout.read())
    err = stderr.read()
    if len(err) != 0:
        logging.error('error occurs: {}'.format(err))
    else:
        logging.info('success')


def _parse_log_to_file(log_dir, task_dag_json_dir):
    for log_filename in os.listdir(log_dir):
        ret = []
        logging.info('parsing {}'.format(log_filename))
        with open(os.path.join(log_dir, log_filename), 'r') as fd:
            for line in fd:
                match = re.search(r'\["MPPTask:<query:\d+,task:\d+> mpp_task_tracing (\{.+\})', line)
                if match is None:
                    continue
                json_str = match.group(1).replace('\\', '')
                try:
                    data = json.loads(json_str)
                except Exception as e:
                    logging.error('failed to load json: {}\n{}'.format(e, json_str))
                    continue
                ret.append(data)
        output_filename = os.path.join(task_dag_json_dir, log_filename + '.task_dag.json')
        logging.info('write to {}'.format(output_filename))
        with open(output_filename, 'wt') as fd:
            json.dump(ret, fd, indent=1)


def _clean_cluster(cluster_name):
    cluster_dir = os.path.join(FLASHPROF_CLUSTER_DIR, cluster_name)
    if os.path.exists(cluster_dir):
        shutil.rmtree(cluster_dir)


def _combine_json_files(json_dir, except_list):
    # TODO: potential optimization, make this stream-ish to reduce memory consumption
    combined = []
    for filename in os.listdir(json_dir):
        if filename in except_list:
            continue
        data = utils.read_file(os.path.join(json_dir, filename))
        # this must success, because it should dumped with this same script
        parsed = json.loads(data)
        combined.extend(parsed)
    with open(os.path.join(json_dir, 'cluster.json'), 'wt') as fd:
        json.dump(combined, fd, indent=1)


def _parse_cluster_log(cluster_name):
    log_dir = os.path.join(FLASHPROF_CLUSTER_DIR, cluster_name, 'log')
    task_dag_json_dir = os.path.join(FLASHPROF_CLUSTER_DIR, cluster_name, 'task_dag', 'json')
    utils.ensure_dir_exist(log_dir)
    utils.ensure_dir_exist(task_dag_json_dir)
    logging.info('parse logs in {}'.format(log_dir))
    _parse_log_to_file(log_dir, task_dag_json_dir)
    _combine_json_files(task_dag_json_dir, ['cluster.json'])


def collect(parser, args):
    _clean_cluster(args.cluster)
    username, ssh_key_file, tiflash_servers = _get_tiup_config(args.cluster)
    log_dir = os.path.join(FLASHPROF_CLUSTER_DIR, args.cluster, 'log')
    task_dag_json_dir = os.path.join(FLASHPROF_CLUSTER_DIR, args.cluster, 'task_dag', 'json')
    utils.ensure_dir_exist(log_dir)
    utils.ensure_dir_exist(task_dag_json_dir)
    for server in tiflash_servers:
        _copy_log_file(server['host'], server['ssh_port'], username, ssh_key_file, server['log_dir'], args.cluster)
    _parse_cluster_log(args.cluster)


def parse(parser, args):
    if args.cluster is None:
        # all clusters
        for cluster_name in os.listdir(FLASHPROF_CLUSTER_DIR):
            _parse_cluster_log(cluster_name)
    else:
        cluster_dir = os.path.join(FLASHPROF_CLUSTER_DIR, args.cluster)
        if not os.path.exists(cluster_dir):
            raise FileNotFoundError('cannot find cluster dir for {}, should be {}'.format(args.cluster, cluster_dir))
        _parse_cluster_log(args.cluster)


def _render_files(json_path, out_dir, type, format):
    '''
    The original json may contain multiple distinct query tso, so we need to split it into multiple input
    with a single tso in each.
    The number of output file(s) is equal to distinct(query_tso) in the original json file.
    '''
    json_data = utils.read_json(json_path)

    # split dataset according to query_tso
    tso_partitioned = defaultdict(lambda: defaultdict(list))  # {tso: {task_id: [INITIALIZING, FINISHED/CANCELLED]}}
    for data in json_data:
        query_tso = data['query_tso']
        task_id = data['task_id']
        tso_partitioned[query_tso][task_id].append(data)

    # only for debugging
    for query_tso, data in tso_partitioned.items():
        utils.ensure_dir_exist(os.path.join(out_dir, '.json_debug'))
        with open(os.path.join(out_dir, '.json_debug', '{}.json'.format(query_tso)), 'wt') as fd:
            json.dump(data, fd, indent=1)

    # overwrite task with INITIALIZING status with CANCELLED/FINISHED
    status_pruned = defaultdict(list)  # {tso: [list of tasks]}
    for query_tso, tasks_map in tso_partitioned.items():
        for task_id, tasks in tasks_map.items():
            if len(tasks) == 1:  # only INITIALIZING
                if tasks[0]['status'] != 'INITIALIZING':
                    raise ValueError(
                        'expect the only task has status INITIALIZING, tso [{}], task_id [{}]'.format(query_tso, task_id))
                status_pruned[query_tso].append(tasks[0])
            elif len(tasks) == 2:  # has INITIALIZING and FINISHED/CANCELLED
                if tasks[0]['status'] == tasks[1]['status'] == 'INITIALIZING':
                    raise ValueError(
                        '2 tasks must not all have INITIALIZING status, tso [{}], task_id [{}]'.format(query_tso, task_id))
                if tasks[0]['status'] != 'INITIALIZING' and tasks[0]['status'] != 'INITIALIZING':
                    raise ValueError(
                        'at least 1 task must have INITIALIZING status, tso [{}], task_id [{}]'.format(query_tso, task_id))
                _task = tasks[0]
                if _task['status'] == 'INITIALIZING':
                    _task = tasks[1]
                status_pruned[query_tso].append(_task)
            else:
                raise ValueError(
                    'expect only 1 or 2 tasks with status, got [{}], tso [{}], task_id [{}]'.format(len(tasks), query_tso, task_id))

    # NOTE: we can parallelize this and utilize all the cpu cores to render
    for query_tso, data in status_pruned.items():
        logging.debug('query_tso [{}], count [{}]'.format(query_tso, len(data)))
        out_path = os.path.join(out_dir, '{}.dot'.format(query_tso))
        logging.info('rendering query_tso [{}]'.format(query_tso))
        if type == RENDER_TYPE.TASK_DAG:
            draw_tasks_dag(data, out_path, format)
        else:
            raise Exception('type {} is not supported yet'.format(type))


def render_one(parser, args):
    _render_files(args.json_file, args.out_dir, args.type, args.format)


def render_cluster(parser, args):
    cluster_dirs = []
    # all clusters
    if args.cluster is None:
        for cluster_name in os.listdir(FLASHPROF_CLUSTER_DIR):
            cluster_dirs.append(os.path.join(FLASHPROF_CLUSTER_DIR, cluster_name))
    else:
        cluster_dirs = [os.path.join(FLASHPROF_CLUSTER_DIR, args.cluster)]

    for cluster_dir in cluster_dirs:
        logging.debug('start rendering for {}'.format(cluster_dir))
        json_path = os.path.join(cluster_dir, 'task_dag', 'json', 'cluster.json')
        out_dir = os.path.join(cluster_dir, 'task_dag', args.format)
        _render_files(json_path, out_dir, RENDER_TYPE.TASK_DAG, args.format)


def default(parser, args):
    parser.print_help()


def cli():
    parser = argparse.ArgumentParser()
    parser.add_argument('--log', type=str, default='info',
                        help='log level, use debug for more detailed logs, default to info')
    parser.set_defaults(func=default)

    subparsers = parser.add_subparsers()

    parser_collect = subparsers.add_parser(
        'collect', help='try to use tiup configuration in this machine for the cluster specified in args')
    parser_collect.add_argument('--cluster', type=str, required=True)
    parser_collect.set_defaults(func=collect)

    parser_render = subparsers.add_parser('render', help='render task dag files into graphic format')
    parser_render.add_argument('--cluster', type=str)
    parser_render.add_argument('--format', type=str, default='png')
    parser_render.set_defaults(func=render_cluster)

    parser_parse = subparsers.add_parser('parse', help='default to parse all clusters\' logs, mainly for debugging')
    parser_parse.add_argument('--cluster', type=str)
    parser_parse.set_defaults(func=parse)

    parser_render_one = subparsers.add_parser('render_one', help='render one file, mainly for debugging')
    parser_render_one.add_argument('--json_file', type=str, required=True)
    parser_render_one.add_argument('--out_dir', type=str, required=True)
    parser_render_one.add_argument('--type', type=RENDER_TYPE, default=RENDER_TYPE.TASK_DAG, choices=list(RENDER_TYPE))
    parser_render_one.add_argument('--format', type=str, default='png')
    parser_render_one.set_defaults(func=render_one)

    args = parser.parse_args(sys.argv[1:])

    numeric_level = getattr(logging, args.log.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: {}'.format(args.log))
    logging.basicConfig(level=numeric_level,
                        format='%(asctime)s.%(msecs)03d %(levelname)s %(message)s', datefmt='%H:%M:%S')
    logging.debug('logging level is set to {}'.format(args.log.upper()))

    args.func(parser, args)

    logging.debug('done')


if __name__ == '__main__':
    cli()



import json
import os


def read_file(filename):
    with open(filename, 'rt') as fd:
        return fd.read()


def read_json(filename):
    with open(filename, 'rt') as fd:
        data = json.load(fd)
    return data


def ensure_dir_exist(dir_name):
    if os.path.exists(dir_name):
        return
    os.makedirs(dir_name)



import json
from collections import defaultdict
from typing import Any, Dict

import graphviz
import pandas as pd

OUTPUT_DIR = 'output'


def draw_test():
    g = graphviz.Graph(comment='The Round Table')
    dot = graphviz.Graph(name='cluster_sub1')
    dot.attr(label='sub1')
    # dot.attr(bgcolor='lightgreen')
    dot.node('A', 'A')
    dot.node('B', 'B')
    dot.node('L', 'L')
    dot.edge('A', 'B')
    dot.edge('A', 'L')
    dot.edge('B', 'L', constraint='false')
    g.subgraph(dot)
    dot = graphviz.Graph(name='cluster_sub2')
    dot.node('A1', 'A1')
    dot.node('B1', 'B1')
    dot.node('L1', 'L1')
    dot.edge('A1', 'B1')
    dot.edge('A1', 'L1')
    dot.edge('B1', 'L1', constraint='false')
    g.subgraph(dot)
    g.render('{}/draw_test'.format(OUTPUT_DIR), format='png')
    print(g.source)


TaskSets = Dict[str, Dict[str, Any]]


def draw(task_sets: TaskSets):
    g = graphviz.Digraph()
    subgraph_index = 0
    for q_id, tasks in task_sets.items():
        q_id = str(q_id)
        s = graphviz.Digraph(name='cluster_'+q_id)
        s.attr(label='q'+q_id)
        print(tasks)
        for t_id, payload in tasks.items():
            parent_ids = payload['parents']
            outputThroughput = payload['outputThroughput']
            t_id_label = 'task{}\noutput throuput: {}'.format(t_id, outputThroughput)
            t_id = 'q{}t{}'.format(subgraph_index, t_id)
            s.node(t_id, t_id_label)
            for parent in parent_ids:
                # skip root node
                if parent == -1:
                    continue
                parent_id = 'q{}t{}'.format(subgraph_index, parent)
                s.edge(t_id, parent_id)
        g.subgraph(s)
        subgraph_index += 1

    print(g.source)
    g.render('{}/query_task'.format(OUTPUT_DIR), format='png')


def parse(df: pd.DataFrame) -> TaskSets:
    task_sets = defaultdict(lambda: {})
    for label, row in df.iterrows():
        queryTso = row['queryTso']
        taskId = row['taskId']
        parents = json.loads(row['upstreamTaskIds'])
        outputThroughput = row['outputThroughput']
        task_sets[queryTso][taskId] = {
            'parents': parents,
            'outputThroughput': outputThroughput
        }
    return dict(task_sets)


def read_csv(filename: str) -> pd.DataFrame:
    df = pd.read_csv(filename)
    df = df[['queryTso', 'taskId', 'upstreamTaskIds', 'outputThroughput']]
    return df


if __name__ == '__main__':
    # draw_test()
    df = read_csv('tracing.csv')
    task_sets = parse(df)
    draw(task_sets)



import json
import logging
from collections import defaultdict
from typing import Any, Dict, List

import graphviz

from utils import read_json

OUTPUT_DIR = 'output'


def _gen_label_executor(detail):
    not_keys = {'type', 'id', 'children', 'sender_target_task_ids', 'receiver_source_task_ids', 'connection_details'}
    return _gen_label(detail, not_keys)


def _gen_label_executor_task(task):
    not_keys = {'executors', 'sender_executor_id'}
    return _gen_label(task, not_keys)


def _gen_label_input_stream(task):
    not_keys = {'id', 'name', 'stat'}
    outter = _gen_label(task, not_keys)
    not_keys_inner = {'timeline'}
    inner = _gen_label(task['stat'], not_keys_inner)
    return outter + inner


def _gen_label_input_stream_task(task):
    not_keys = {'input_streams'}
    return _gen_label(task, not_keys)


def _gen_label(data: Dict, not_keys, join_char='\l'):
    labels = []
    for k, v in data.items():
        if k not in not_keys:
            labels.append('{}: {}'.format(k, v))
    return join_char.join(labels) + join_char


# translate input json list [task1, task2] to {id1: task1} for better performance and code structure
def _trans_list_to_id_map(l: List) -> Dict[int, Any]:
    ret = {}
    for i in l:
        ret[i['id']] = i
    return ret


class InputStreamTaskGraph:
    def __init__(self, task):
        self._g = graphviz.Digraph(name='cluster_'+str(task['task_id']), comment='input stream graph')
        self._g.attr(label=_gen_label_input_stream_task(task), labeljust='l', labelloc='b')
        self._task_id = task['task_id']
        self._input_streams = _trans_list_to_id_map(task['input_streams'])

    @property
    def g(self):
        return self._g

    def draw_input_streams(self):
        self._draw_nodes()
        self._draw_edges()

    def _draw_nodes(self):
        for id, s in self._input_streams.items():
            label = '{}_{}\l{}'.format(s['name'], id, _gen_label_input_stream(s))
            self._g.node(self.get_node_id(id), label, shape='box')

    def _draw_edges(self):
        for id, s in self._input_streams.items():
            for child_id in s['children']:
                self._g.edge(self.get_node_id(child_id), self.get_node_id(id))

    def get_node_id(self, node_id):
        return '{}-{}'.format(self._task_id, node_id)


class InputStreamGraph:
    def __init__(self):
        self._g = graphviz.Digraph(comment='main graph')
        self._g.attr(rankdir='BT')

    def addTaskGraph(self, task_graph: InputStreamTaskGraph):
        self._g.subgraph(task_graph.g)

    def render(self):
        path = '{}/input_stream_task.dot'.format(OUTPUT_DIR)
        self._g.render(path, format='png')


class TaskGraph:
    def __init__(self, task):
        self._task = task
        self._executors = _trans_list_to_id_map(task['executors'])
        self._task_id = task['task_id']
        self._receiver_sources: Dict[int, List[int]] = {}
        self._g = graphviz.Digraph(name='cluster_'+str(task['task_id']), comment='task')
        if self._task['status'] == 'FINISHED' and self._task['error_message'] == '':
            self._g.attr(label=_gen_label_executor_task(task), labeljust='l', labelloc='b', style='solid')
        else:
            self._g.attr(label=_gen_label_executor_task(task), labeljust='l',
                         labelloc='b', style='solid', color='red', penwidth='3')

    @property
    def g(self):
        return self._g

    @property
    def id(self):
        return self._task_id

    @property
    def sender_executor_id(self):
        return self._task['sender_executor_id']

    @property
    def sender_executor(self):
        sender_executor_id = self._task['sender_executor_id']
        return self._executors[sender_executor_id]

    @property
    def receiver_sources(self):
        return self._receiver_sources

    def draw_executors(self):
        self._draw_executor_nodes()
        self._draw_executor_edges()
        self._collect_receiver_sources()

    def _collect_receiver_sources(self):
        for eid, e in self._executors.items():
            if 'receiver_source_task_ids' in e:
                self._receiver_sources[eid] = e['receiver_source_task_ids']

    def _draw_executor_nodes(self):
        for eid, e in self._executors.items():
            label = '{}_{}\l{}'.format(e['type'], e['id'], _gen_label_executor(e))
            self._g.node(self.get_node_id(eid), label, shape='box')

    def _draw_executor_edges(self):
        for eid, e in self._executors.items():
            for child_id in e['children']:
                self._g.edge(self.get_node_id(child_id), self.get_node_id(eid), weight='0')

    def get_node_id(self, node_id):
        return '{}-{}'.format(self._task_id, node_id)


class Graph:
    def __init__(self):
        self._g = graphviz.Digraph(comment='main')
        self._g.attr(rankdir='BT', splines='line')
        self._task_graphs: Dict[int, TaskGraph] = {}
        self._stages: Dict[int, List[TaskGraph]] = defaultdict(list)

    def addTaskGraph(self, task_graph: TaskGraph):
        # self._g.subgraph(task_graph.g)
        self._stages[task_graph.sender_executor_id].append(task_graph)
        self._task_graphs[task_graph.id] = task_graph

    def _draw_stages(self):
        for sender_executor_id, task_graphs in self._stages.items():
            stage_g = graphviz.Digraph(name='cluster_{}'.format(sender_executor_id), comment='stage')
            stage_g.attr(label='stage', labeljust='c', labelloc='b', style='dashed', rank='same')
            for task_graph in task_graphs:
                stage_g.subgraph(task_graph.g)
            self._g.subgraph(stage_g)

    # draw executor references corssing task boundaries
    def _draw_task_references(self):
        for receiver_task_id, receiver_task_graph in self._task_graphs.items():
            for receiver_executor_id, sender_task_ids in receiver_task_graph.receiver_sources.items():
                receiver_node_id = receiver_task_graph.get_node_id(receiver_executor_id)
                for sender_task_id in sender_task_ids:
                    if sender_task_id not in self._task_graphs:
                        logging.error('failed to get source task with task_id [{}] for receiver task [{}]'.format(
                            sender_task_id, receiver_task_id))
                        continue
                    sender_task_graph = self._task_graphs[sender_task_id]
                    sender_executor_id = sender_task_graph.sender_executor_id
                    sender_node_id = sender_task_graph.get_node_id(sender_executor_id)
                    self._g.edge(sender_node_id, receiver_node_id, color='red', style='dashed')

    def render(self, filename, format):
        self._draw_stages()
        self._draw_task_references()
        self._g.render(filename, format=format)


def draw_tasks_dag(data, filename, format='png'):
    if filename is None:
        filename = '{}/query_task.dot'.format(OUTPUT_DIR)
    graph = Graph()
    for task in data:
        task_graph = TaskGraph(task)
        task_graph.draw_executors()
        graph.addTaskGraph(task_graph)
    graph.render(filename, format)


def draw_input_streams():
    data = read_json('/Users/dragonly/Downloads/test-input-streams.json')
    graph = InputStreamGraph()
    for task in data:
        task_graph = InputStreamTaskGraph(task)
        task_graph.draw_input_streams()
        graph.addTaskGraph(task_graph)
    graph.render()


# TODO: query tso should be fixed
def draw_input_streams_timeline():
    data = read_json('/Users/dragonly/Downloads/test-input-streams(1).json')
    events = []
    task_id_set = set()

    # first round, collect metadata events (M)
    # for the moment, we use ui.perfetto.dev to draw the timeline
    # using task as process/stream as thread is good for drawing
    for task in data:
        task_id = task['query_tso']
        if task_id not in task_id_set:
            task_id_set.add(task_id)
            events.append({
                'pid': task_id,
                'ph': 'M',
                'name': 'process_name',
                'args': {'name': 'task {}'.format(task_id)}
            })
        input_streams = task['input_streams']
        for stream in input_streams:
            stream_id = stream['id']
            events.append({
                'pid': task_id,
                'tid': stream_id,
                'ph': 'M',
                'name': 'thread_name',
                'args': {'name': 'stream {}'.format(stream_id)}
            })

    # second round, collect counter events
    for task in data:
        task_id = task['query_tso']
        input_streams = task['input_streams']
        for stream in input_streams:
            stream_id = stream['id']
            name = '[{}] {}({})'.format(stream_id, stream['name'], stream['executor'])
            stats = stream['stat']
            timeline_push = stats['timeline']['push']
            timeline_pull = stats['timeline']['pull']
            timeline_self = stats['timeline']['self']
            events_push = _gen_counter_events(task_id, stream_id, name, timeline_push, timeline_pull, timeline_self)
            events.extend(events_push)

    with open('output/timeline.json', 'w') as fd:
        json.dump(events, fd)

# currently we assume that each


def _gen_counter_events(pid, tid, name, push, pull, self):
    ret = []
    ts = 1
    for c1, c2, c3 in zip(push, pull, self):
        ret.append({
            'pid': pid,
            'tid': tid,
            'ph': 'C',
            'name': name,
            'ts': ts,
            'args': {'push': c1, 'pull': c2, 'self': c3}
        })
        ts += 100*1000  # microsecond
    return ret


if __name__ == '__main__':
    # TODO: consider multiple query tso
    draw_tasks_dag(read_json('/Users/dragonly/Downloads/multi_machine_mpp_task_tracing.json'))
    # draw_input_streams()
    # draw_input_streams_timeline()



from setuptools import setup


def read_file(filename):
    with open(filename, 'r') as fd:
        return fd.read()


# see examples in https://docs.python.org/3/distutils/examples.html
setup(
    name='flashprof',
    version='0.0.8',
    description='a tool that collects and visualizes TiFlash runtime infomation',
    long_description=read_file('README.md'),
    long_description_content_type='text/markdown',
    author='dragonly',
    author_email='liyilongko@gmail.com',
    url='https://github.com/dragonly/tiflash_mpp_profiler',
    license=read_file('LICENSE'),
    package_dir={'': 'src'},
    packages=[''],
    entry_points={
        'console_scripts': [
            'flashprof = main:cli'
        ]
    },
    install_requires=[
        'graphviz',
        'paramiko',
        'pyyaml'
    ]
)
